// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI6C500K/24DSI6C500K_1.x.x.x_GSC_DN/driver/device.c $
// $Rev: 54804 $
// $Date: 2024-07-08 08:07:15 -0500 (Mon, 08 Jul 2024) $

// 24DSI6C500K: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model			Vendor	Device	SubVen	SubDev	type
	{ "24DSI6C500K",	0x10B5, 0x9056, 0x10B5, 0x3551,	GSC_DEV_TYPE_24DSI6C500K	},
	{ NULL }
};



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	dev->cache.channels_max	= 6;

	switch (dev->cache.gsc_bcfgr_32 & 0x70000)
	{
		default:

			printk(	"%s, #%d: Board Config Register 0x%lX, D16-D18 unrecognized\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);

			FALLTHROUGH;	/* fall through */

		case 0x00000:	dev->cache.channel_qty	= 6;	break;
		case 0x10000:	dev->cache.channel_qty	= 4;	break;
		case 0x20000:	dev->cache.channel_qty	= 2;	break;
	}
}



//*****************************************************************************
static void _filter_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_bcfgr_32 & 0x180000)
	{
		default:

			printk(	"%s, #%d: Board Config Register 0x%lX, D19-D20 unrecognized\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);
			dev->cache.filter_freq	= -1;
			break;

		case 0x000000:	dev->cache.filter_freq	= 2000000L;	break;
		case 0x080000:	dev->cache.filter_freq	= 0;		break;
	}
}



//*****************************************************************************
static void _range_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_bcfgr_32 & 0x600000)
	{
		default:
		case 0x000000:	dev->cache.voltage_range	= DSI6C500K_QUERY_V_RANGE_6;	break;
		case 0x200000:	dev->cache.voltage_range	= DSI6C500K_QUERY_V_RANGE_10;	break;
		case 0x400000:	dev->cache.voltage_range	= DSI6C500K_QUERY_V_RANGE_5;	break;
		case 0x600000:	dev->cache.voltage_range	= DSI6C500K_QUERY_V_RANGE_2_5;	break;
	}
}



//*****************************************************************************
static void _sync_src_compute(dev_data_t* dev)
{
	if ((dev->cache.gsc_bcfgr_32 & 0xFFF) >= 0x002)
		dev->cache.sync_src	= 1;
	else
		dev->cache.sync_src	= 0;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(DSI6C500K_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_avr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_AVR));
		dev->vaddr.gsc_bbsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BBSR));
		dev->vaddr.gsc_bcfgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BCFGR));
		dev->vaddr.gsc_bctlr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BCTLR));
		dev->vaddr.gsc_bttr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BTTR));
		dev->vaddr.gsc_bufcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BUFCR));
		dev->vaddr.gsc_bufsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_BUFSR));
		dev->vaddr.gsc_csar_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_CSAR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_DIOPR));
		dev->vaddr.gsc_idbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_IDBR));
		dev->vaddr.gsc_rcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_RCR));
		dev->vaddr.gsc_rdr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI6C500K_GSC_RDR));

		// Data cache initialization.
		dev->cache.gsc_bcfgr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcfgr_32);

		dev->model					= dev_id_list[0].model;
		dev->cache.autocal_ms		= 11000;
		dev->cache.fclk_default		= 50000000L;
		dev->cache.fgen_max			= 40000000L;
		dev->cache.fgen_min			= 20000000L;
		dev->cache.fifo_size		= _256K;
		dev->cache.fref_default		= 32768000L;
		dev->cache.fsamp_max		= 500000;
		dev->cache.fsamp_min		= 9766;
		dev->cache.initialize_ms	= 20;
		dev->cache.ndiv_max			= 4;
		dev->cache.ndiv_min			= 1;
		dev->cache.nref_max			= 300;
		dev->cache.nref_min			= 25;
		dev->cache.nvco_max			= 300;
		dev->cache.nvco_min			= 25;

		_channels_compute(dev);
		_filter_compute(dev);
		_range_compute(dev);
		_sync_src_compute(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_DMDMA_READ;
		ret	= gsc_dma_create(dev, dma, dma);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


